from datetime import date, timedelta

from django.conf import settings
from django.db import models
from django.db.models import Max
from django.utils import timezone
from django.utils.translation import gettext_lazy as _

from user_payments.models import LineItem

from .utils import recurring


class SubscriptionManager(models.Manager):

    def create_periods(self):
        for subscription in self.all():
            subscription.create_periods()


class Subscription(models.Model):
    user = models.ForeignKey(
        settings.AUTH_USER_MODEL,
        on_delete=models.PROTECT,
        related_name="user_subscriptions",
        verbose_name=_("user"),
    )
    created_at = models.DateTimeField(_("created at"), default=timezone.now)
    title = models.CharField(_("title"), max_length=200)
    starts_at = models.DateTimeField(_("starts at"))
    ends_at = models.DateTimeField(_("ends at"), blank=True, null=True)
    periodicity = models.CharField(
        _("periodicity"),
        max_length=20,
        choices=[
            ("yearly", _("yearly")),
            ("monthly", _("monthly")),
            ("weekly", _("weekly")),
            ("manually", _("manually")),
        ],
    )
    amount = models.DecimalField(_("amount"), max_digits=10, decimal_places=2)

    renew_automatically = models.BooleanField(_("renew automatically"))
    paid_until = models.DateTimeField(_("paid until"), blank=True)

    objects = SubscriptionManager()

    class Meta:
        verbose_name = _("subscription")
        verbose_name_plural = _("subscriptions")

    def __str__(self):
        return self.title

    def save(self, *args, **kwargs):
        if not self.paid_until:
            self.paid_until = self.starts_at or timezone.now()
        super().save(*args, **kwargs)

    save.alters_data = True

    def update_paid_until(self, save=True):
        self.paid_until = self.periods.paid().aggregate(m=Max("ends_at"))[
            "m"
        ] or self.paid_until
        if save:
            self.save()

    @property
    def is_active(self):
        # TODO grace period?
        return self.starts_at <= timezone.now() <= self.paid_until

    def create_periods(self, *, until=None):
        """
        Create period instances for this subscription, up to either today or
        the end of the subscription, whichever date is earlier.
        """
        end = until or date.today()
        if self.ends_at:
            end = min(self.ends_at, end)
        days = recurring(self.starts_at, self.periodicity)
        this_start = next(days)

        periods = list(self.periods.all())

        existing = set(p.starts_at for p in periods)
        while True:
            next_start = next(days)
            if this_start not in existing:
                p, _created = self.periods.get_or_create(
                    starts_at=this_start, ends_at=next_start - timedelta(days=1)
                )
                periods.append(p)
            this_start = next_start

            # TODO This might be a good place to already generate periods
            # for the near future to inform users that a payment will soon
            # be due.

            if this_start >= end:
                break

        return periods

    create_periods.alters_data = True


class SubscriptionPeriodManager(models.Manager):

    def paid(self):
        """
        Return subscription periods that have been paid for.
        """
        return self.filter(line_item__payment__charged_at__isnull=False)

    def create_line_items(self):
        # Not really a good QuerySet method, should be a manager method, but
        # whatever...
        for period in self.filter(line_item__isnull=True):
            period.create_line_item()


class SubscriptionPeriod(models.Model):
    subscription = models.ForeignKey(
        Subscription,
        on_delete=models.CASCADE,
        related_name="periods",
        verbose_name=_("subscription"),
    )
    starts_at = models.DateField(_("starts at"))
    ends_at = models.DateField(_("ends at"))
    line_item = models.OneToOneField(
        LineItem,
        on_delete=models.CASCADE,
        blank=True,
        null=True,
        verbose_name=_("line item"),
    )

    objects = SubscriptionPeriodManager()

    class Meta:
        unique_together = (("subscription", "starts_at"),)
        verbose_name = _("subscription period")
        verbose_name_plural = _("subscription periods")

    def __str__(self):
        return "%s (%s - %s)" % (self.subscription, self.starts_at, self.ends_at)

    def create_line_item(self):
        """
        Create a user payments line item for this subscription period.
        """
        # TODO Maybe create periods and line items early,
        # e.g. 7 days before new period begins?
        self.line_item = LineItem.objects.create(
            user=self.subscription.user,
            title=str(self),
            amount=self.subscription.amount,
        )
        self.save()

    create_line_item.alters_data = True



# Generated by Django 2.0.5 on 2018-05-15 11:51

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [("user_subscriptions", "0001_initial")]

    operations = [
        migrations.AlterField(
            model_name="subscription",
            name="periodicity",
            field=models.CharField(
                choices=[
                    ("yearly", "yearly"), ("monthly", "monthly"), ("weekly", "weekly")
                ],
                max_length=20,
                verbose_name="periodicity",
            ),
        )
    ]



# Generated by Django 2.0.5 on 2018-05-15 12:04

from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    dependencies = [
        ("user_payments", "0001_initial"),
        ("user_subscriptions", "0002_auto_20180515_1351"),
    ]

    operations = [
        migrations.AddField(
            model_name="subscriptionperiod",
            name="line_item",
            field=models.OneToOneField(
                blank=True,
                null=True,
                on_delete=django.db.models.deletion.CASCADE,
                related_name="+",
                to="user_payments.LineItem",
                verbose_name="line item",
            ),
        )
    ]



import logging

from django.utils import timezone

import stripe
from stripe_subscriptions.models import Customer
from user_payments.models import Payment


logger = logging.getLogger(__name__)


class StripeSubscriptionsProcessor:

    def process(self):
        for customer in Customer.objects.select_related("user"):
            payment = Payment.objects.create_pending(user=customer.user)
            if not payment:
                continue

            try:
                # Really do this right away?
                charge = stripe.Charge.create(
                    customer=customer.customer_id,
                    amount=payment.instance.amount_cents,
                    currency="CHF",
                    idempotency_key=payment.id.hex,
                )
            except stripe.CardError as exc:
                # TODO do a thing...
                logger.exception(exc)

                # TODO Probably should also send a mail?
                payment.lineitems.update(payment=None)
                payment.delete()

            else:
                # TODO check for more errors?
                payment.charged_at = timezone.now()
                payment.transaction = str(charge)
                payment.save()

            # FIXME sender? request?
            # post_charge.sender(sender=self, payment=instance, request=None)



# Generated by Django 2.0.5 on 2018-05-17 12:35

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [("user_subscriptions", "0003_subscriptionperiod_line_item")]

    operations = [
        migrations.RenameField(
            model_name="subscription", old_name="is_active", new_name="_is_active"
        ),
        migrations.AddField(
            model_name="subscription",
            name="_is_automatically_renewed",
            field=models.BooleanField(
                default=False, verbose_name="is automatically renewed"
            ),
            preserve_default=False,
        ),
        migrations.AlterField(
            model_name="subscription",
            name="ends_at",
            field=models.DateTimeField(blank=True, null=True, verbose_name="ends at"),
        ),
        migrations.AlterField(
            model_name="subscription",
            name="periodicity",
            field=models.CharField(
                choices=[
                    ("yearly", "yearly"),
                    ("monthly", "monthly"),
                    ("weekly", "weekly"),
                    ("manually", "manually"),
                ],
                max_length=20,
                verbose_name="periodicity",
            ),
        ),
        migrations.AlterField(
            model_name="subscription",
            name="starts_at",
            field=models.DateTimeField(verbose_name="starts at"),
        ),
    ]



from django.conf import settings
from django.db import models, transaction
from django.utils import timezone
from django.utils.translation import gettext, gettext_lazy as _

from mooch.models import Payment as AbstractPayment


class PaymentManager(models.Manager):

    def create_pending(self, *, user):
        """
        Create an unpaid payment instance with all line items for the given
        user that have not been bound to a payment instance yet.

        Returns ``None`` if there are no unbound line items for the given user.
        """
        # XXX payment__isnull | payment__charged_at__isnull?
        with transaction.atomic():
            pending = LineItem.objects.filter(user=user, payment__isnull=True)
            if not len(pending):
                return None

            payment = self.create(
                user=user, amount=sum((item.amount for item in pending), 0)
            )
            pending.update(payment=payment)
            return payment


class Payment(AbstractPayment):
    user = models.ForeignKey(
        settings.AUTH_USER_MODEL,
        on_delete=models.PROTECT,
        related_name="user_payments",
        verbose_name=_("user"),
    )

    objects = PaymentManager()

    def __str__(self):
        if self.charged_at:
            return gettext("Payment of %s") % self.amount
        return gettext("Pending payment of %s") % self.amount

    def save(self, *args, **kwargs):
        if not self.email:
            self.email = self.user.email
        super().save(*args, **kwargs)

    save.alters_data = True


class LineItem(models.Model):
    """
    Individual line items may be created directly using the manager method. A
    minimal example follows::

        @login_required
        def some_view(request):
            # This request costs 5 cents!
            LineItem.objects.create(
                user=request.user,
                amount=Decimal('0.05'),
                title='Request to some_view',
            )
            # Rest of view

    If you already have a payment instance at hand you may pass it as well.
    """

    user = models.ForeignKey(
        settings.AUTH_USER_MODEL,
        on_delete=models.PROTECT,
        related_name="user_lineitems",
        verbose_name=_("user"),
    )
    created_at = models.DateTimeField(_("created at"), default=timezone.now)
    title = models.CharField(_("title"), max_length=200)
    amount = models.DecimalField(_("amount"), max_digits=10, decimal_places=2)

    payment = models.ForeignKey(
        Payment,
        on_delete=models.PROTECT,
        blank=True,
        null=True,
        related_name="lineitems",
        verbose_name=_("payment"),
    )

    class Meta:
        ordering = ["-created_at"]
        verbose_name = _("line item")
        verbose_name_plural = _("line items")

    def __str__(self):
        return self.title



from django.contrib import admin

from . import models


@admin.register(models.Subscription)
class SubscriptionAdmin(admin.ModelAdmin):
    list_display = (
        "user",
        "title",
        "starts_at",
        "ends_at",
        "periodicity",
        "amount",
        "paid_until",
        "renew_automatically",
    )
    list_filter = ("renew_automatically",)
    radio_fields = {"periodicity": admin.HORIZONTAL}
    raw_id_fields = ("user",)
    search_fields = ("title",)


@admin.register(models.SubscriptionPeriod)
class SubscriptionPeriodAdmin(admin.ModelAdmin):
    list_display = ("subscription", "starts_at", "ends_at", "line_item")
    raw_id_fields = ("subscription", "line_item")



from django.core.management.base import BaseCommand

from user_subscriptions.models import Subscription, SubscriptionPeriod


class Command(BaseCommand):
    help = "Generate subscription periods and user payments line items"

    def handle(self, **options):
        Subscription.objects.create_periods()
        SubscriptionPeriod.objects.create_line_items()



# Generated by Django 2.0.5 on 2018-05-15 11:38

from django.conf import settings
from django.db import migrations, models
import django.db.models.deletion
import django.utils.timezone


class Migration(migrations.Migration):

    initial = True

    dependencies = [migrations.swappable_dependency(settings.AUTH_USER_MODEL)]

    operations = [
        migrations.CreateModel(
            name="Subscription",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                (
                    "created_at",
                    models.DateTimeField(
                        default=django.utils.timezone.now, verbose_name="created at"
                    ),
                ),
                ("title", models.CharField(max_length=200, verbose_name="title")),
                ("starts_at", models.DateField(verbose_name="starts at")),
                (
                    "ends_at",
                    models.DateField(blank=True, null=True, verbose_name="ends at"),
                ),
                (
                    "periodicity",
                    models.CharField(
                        choices=[("monthly", "monthly"), ("weekly", "weekly")],
                        max_length=20,
                        verbose_name="periodicity",
                    ),
                ),
                (
                    "amount",
                    models.DecimalField(
                        decimal_places=2, max_digits=10, verbose_name="amount"
                    ),
                ),
                ("is_active", models.BooleanField(verbose_name="is active")),
                (
                    "user",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.PROTECT,
                        related_name="+",
                        to=settings.AUTH_USER_MODEL,
                        verbose_name="user",
                    ),
                ),
            ],
            options={
                "verbose_name": "subscription", "verbose_name_plural": "subscriptions"
            },
        ),
        migrations.CreateModel(
            name="SubscriptionPeriod",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("starts_at", models.DateField(verbose_name="starts at")),
                ("ends_at", models.DateField(verbose_name="ends at")),
                (
                    "subscription",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="periods",
                        to="user_subscriptions.Subscription",
                        verbose_name="subscription",
                    ),
                ),
            ],
            options={
                "verbose_name": "subscription period",
                "verbose_name_plural": "subscription periods",
            },
        ),
        migrations.AlterUniqueTogether(
            name="subscriptionperiod", unique_together={("subscription", "starts_at")}
        ),
    ]



# Generated by Django 2.0.5 on 2018-05-17 07:32

from django.conf import settings
from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    dependencies = [("user_subscriptions", "0005_auto_20180517_1441")]

    operations = [
        migrations.AlterField(
            model_name="subscription",
            name="user",
            field=models.ForeignKey(
                on_delete=django.db.models.deletion.PROTECT,
                related_name="user_subscriptions",
                to=settings.AUTH_USER_MODEL,
                verbose_name="user",
            ),
        ),
        migrations.AlterField(
            model_name="subscriptionperiod",
            name="line_item",
            field=models.OneToOneField(
                blank=True,
                null=True,
                on_delete=django.db.models.deletion.CASCADE,
                to="user_payments.LineItem",
                verbose_name="line item",
            ),
        ),
    ]



# Generated by Django 2.0.5 on 2018-05-17 12:41

from django.db import migrations, models
import django.utils.timezone


class Migration(migrations.Migration):

    dependencies = [("user_subscriptions", "0004_auto_20180517_1435")]

    operations = [
        migrations.RemoveField(model_name="subscription", name="_is_active"),
        migrations.RemoveField(
            model_name="subscription", name="_is_automatically_renewed"
        ),
        migrations.AddField(
            model_name="subscription",
            name="paid_until",
            field=models.DateTimeField(
                default=django.utils.timezone.now, verbose_name="paid until"
            ),
            preserve_default=False,
        ),
        migrations.AddField(
            model_name="subscription",
            name="renew_automatically",
            field=models.BooleanField(
                default=False, verbose_name="renew automatically"
            ),
            preserve_default=False,
        ),
    ]
